ConductorWindow {
	var
		<>numPlayers,
		<>simulation,
		<>synths, // for simulation
		<>button,
		<>timing,
		<>netAddr,
		<>network,
		<>playerAddrs,
		<>oscResps,
		<>window;
		
	*new {
		arg
			numPlayers = 8,
			simulation = false,
			duration = 900; // default: 900 seconds, i.e. 15 minutes
			
		^super.newCopyArgs(numPlayers, simulation).init(duration);
	}
	
	init {
		arg duration;
	
		if(
			simulation,
			{
				synths = Array.newClear(8);
				8.do {
					arg i;
					var pan = 2 / 7 * i - 1;
					
					synths[i] = ImpulseSynth(
						simulation: true,
						pan: pan
					)
				}
			}
		);
			
		timing = Timing(duration);

		/* Network set up. */
		netAddr = NetAddr("0.local", 57120); // for sending messages

		try {
			this.init_osc_responders(numPlayers);
			network = true;
		}
		{
			"WARNING: ConductorWindow couldn't find all the players in the network.".postln;
			network = false;
		};
		
		/* Window set up. */
		window = Window.new(
			"ConductorWindow",
			Rect(320, 0, 320, 180)
		);
		
		window.view.background = Color.black;
		window.front;
		CmdPeriod	.add({window.close});		
		
		/* Button set up.  The button turns the clock on and off. */
		button = Button(window, window.view.bounds.insetBy(10));
		
		button.states = [
			["Start clock", Color.green, Color.black],
			["Stop clock", Color.red, Color.black]
		];	
		
		button.action = {
			arg i;
			i.value.switch(
				1, {this.start_clock},
				0, {this.stop_clock}
			);
		}
		

	}
	
	init_osc_responders {
		arg numPlayers;

		// a NetAddr and OSCresponderNode for each player.
		// Players are "0.local", "1.local", "2.local", etc.

		playerAddrs = Array.newClear(numPlayers);
		oscResps = Array.newClear(numPlayers);

		numPlayers.do {
			arg i;
			
			playerAddrs[i] = NetAddr(
				("%.local").format(i.asString),
				57120
			);
			
			oscResps[i] = OSCresponderNode(
				playerAddrs[i],
				'/conductor',
				{
					arg
						time,
						resp,
						msg;
						
					this.parse(i, msg[1]);
				}
			).add;
			
			// Remove the OSC responder when done.
			CmdPeriod	.add({oscResps[i].remove});
			
		}
	}
	
	set_clocks {
		arg time; // a string, e.g. "12:34"

		("Setting clocks to " ++ time).postln;


/*		
		// Public clock
		netAddr.sendBundle(
			0,
			['/publicMsgWindow', \set_clock, time]
		);

		// 110209: public clock nixed following Yoni's suggestion
		
*/		
		// Private clocks
		if(
			network,
			{
				playerAddrs.do{
					arg i;
					i.sendBundle(
						0,
						['/msgWindow', \set_clock, time]
					)
				}
			}
		);
	}
		
	start_clock {
		arg offset = 0;
	
	
		// Change the button state.  This happens automatically if the user
		// clicks on the button, but not if the `start_clock' method is evaluated
		// directly.
		{button.value = 1}.defer;

		timing.start(offset);
				
		// Every second, update all the MsgWindow clocks.
		SystemClock.sched(
			0,
			{
				this.set_clocks(timing.now.to_time_string.to_rounded_time);
				// Wait one second.
				1
			}
		);

		// Advance the private MsgWindow timekeepers five times a second.
		SystemClock.sched(
			0,
			{
				this.advance_timekeepers;
				0.2
			}
		);

		// Stop the clocks when the duration elapses.
		SystemClock.sched(
			timing.duration - offset,
			{
				this.stop_clock;
			}
		);
	}

	stop_clock {
		{button.value = 0}.defer;
		timing.stop;
		this.set_clocks("--:--");
		SystemClock.clear;
		this.reset_timekeepers;
		
		// Turn down the synths.		
		numPlayers.do{
			arg i;
			this.set_synth(i, \amp, 0);
		};
	}
	
	parse {
		arg
			player,
			msg;
			
		var
			eventTime,
			parser,
			groupMsg,
			publicMsg;

		parser = Parser(msg.asString);

		parser.dict.postln;

		if(parser.dict.at("t") != nil, {
			eventTime = parser.dict.at("t");
			if(timing.valid_time(eventTime), {

				/* Sound events. */

				[\amp, \impulseFreq].do{
					arg i;
					var value;
					
					value = parser.dict.at(i);

					if(
						value != nil,
						{
						
							SystemClock.schedAbs(
								timing.startTime + eventTime,
								{
									this.set_synth(player, i, value);
									this.update_freq_tables(player, i, value);
								}
							);
							
						}
					);

				};
				
				// If something was scheduled, reply positively to the sender
				if(
					not(simulation),
					{
						parser.dict.at(\amp).postln;
						parser.dict.at(\freq).postln;
					
						if(
							parser.dict.at(\amp) == 1,
							{
								this.send_private_msg(
									" ", // blank prompt (no player number)
									player,
									"%: Synth on.".format(eventTime.to_time_string)
								)
							}
						);
						if(
							parser.dict.at(\amp) == 0,
							{
								this.send_private_msg(
									" ", // blank prompt (no player number)
									player,
									"%: Synth off.".format(eventTime.to_time_string)								)
							}
						);
						if(
							parser.dict.at(\impulseFreq) != nil,
							{
								this.send_private_msg(
									" ", // blank prompt (no player number)
									player,
									"%: Synth set to % Hz.".format(
										eventTime.to_time_string,
										parser.dict.at(\impulseFreq).asString
									)
								)
							}
						);
					}
				);
				
				// Display the message on the public screen when it's executed.
				AppClock.sched(
					timing.startTime + eventTime - thisThread.seconds,
					{
						this.send_public_msg(player, msg);
					}
				);
			},
			// Interpret malformed messages as messages for the group.

			{
				this.send_group_msg(player, msg);
				
			})
		},
		{
			this.send_group_msg(player, msg);

		});

	}

	parse_msg_list {
		arg
			playerNum,
			msgList;

		msgList.do {
			arg msg;
			this.parse(playerNum, msg)
		}			
	}

	parse_msg_lists {
		arg msgLists;
		
		msgLists.do {
			arg msgList, playerNum;
			this.parse_msg_list(playerNum, msgList)
		}
	}

	time_bounds_error {
		^"ERROR: time out of bounds."
	}

	time_format_error {
		^"ERROR: bad time format, or no time found in the message."
	}	

	advance_timekeepers {
		// private
		playerAddrs.do{
			arg i;	
			i.sendBundle(
				0,
				['/msgWindow', \advance_timekeeper]
			);
		};
		
		// public
/*		netAddr.sendBundle(
			0,
			['/publicMsgWindow', \advance_timekeeper]
		);

		// 110208: no public clock

*/		
	
	}

	reset_timekeepers {
		// private
		playerAddrs.do{
			arg i;
			i.sendBundle(
				0,
				['/msgWindow', \reset_timekeeper]
			);
		};
		
		// public
/*		netAddr.sendBundle(
			0,
			['/publicMsgWindow', \reset_timekeeper]
		);

		// 110208: no public clock
*/		
	}

	send_group_msg {
		arg fromPlayer, msg;
		
		("Sending group message: " ++ msg).postln;

		numPlayers.do{
			arg toPlayer;
			this.send_private_msg(fromPlayer, toPlayer, msg);
		};
	}

	send_private_msg {
		arg
			fromPlayer,
			toPlayer,
			msg;
			
		("Sending private message: " ++ msg).postln;
		
		playerAddrs[toPlayer].sendBundle(
			0,
			['/msgWindow', \add_msg, fromPlayer, msg]
		);
	}

	send_public_msg {
		arg
			fromPlayer,
			msg;
		
		("Sending public message: " ++ msg).postln;
		
		netAddr.sendBundle(
			0,
			['/publicMsgWindow', \add_msg, fromPlayer, msg]
		);	
	}
	
	set_synth {
		arg
			playerNum,
			parameter,
			value;

		// For simulations, send a value to one of the associated synths.
		if(
			simulation,
			{
				parameter.switch(
					\amp, { synths[playerNum].set_amp(value) },
					\impulseFreq, { synths[playerNum].set_freq(value) }
				);			
			},
			{
				playerAddrs[playerNum].sendBundle(
					0,
					['/synth', parameter, value]
				);
			}
		);

	}
	
	update_freq_tables {
		arg
			playerNum,
			parameter,
			value;

		netAddr.sendBundle(
			0,
			['/freqTables', playerNum, parameter, value]
		);
	}	
	
}